/****************************************
 *  Bluescrn's 60fps Bloom Demo
 *
 *	based on 30fps Two Pass 3d Test - Author: eli
 *  which was based on the DrunkenCoders demos by Dovoto
 *
 *  Uses 3 VRAM banks - triple buffering is required to make this work
 *
 *  The final output you see is an 2-frame-old unmodified capture of the 3D scene, on BG2
 *  with a blurred monochrome half-resolution version of the same image blended on top with BG3
 *  The 'live 3D' layer output is hidden below BG2+3, being captured
 *
 *  The C version is now fast enough to manage 60fps, but really
 *  need rewriting in ARM assembler for best results...
 *
 ****************************************/
 

#define	REG_CAPTURE		(*(vu32*)0x04000064)


// include your ndslib
#include <nds.h>
#include <nds/arm9/console.h> 
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <nds/arm9/trig_lut.h>
//needed to load pcx files
#include <nds/arm9/image.h>

typedef struct
{
	int parent;
   m4x4 matrix;
}
ModelBone;

typedef struct
{
	int bone_cnt;
	int* bone_ofs;
	int dlist_cnt;
	int* dlist_ofs;
}
ModelHeader;

#include "ship1_texture_pcx.h"
#include "ship1_model_bin.h"

void SetRegCapture(bool enable, uint8 srcBlend, uint8 destBlend, uint8 bank, uint8 offset, uint8 size, uint8 source, uint8 srcOffset);
int LoadGLTextures();
int DrawShip();

int	xrot;				// X Rotation ( NEW )
int	yrot;				// Y Rotation ( NEW )
int	zrot;				// Z Rotation ( NEW )

int	texture[1];		// Storage For One Texture ( NEW )
ModelHeader* modelPtr;
m4x4 objectMatrix;

// third hand info posted to gbadev.org
/*
capture register is 0x04000064 
ram blend EVA is bits 0..4 
graphics blend evb is 8..12 
vram bank select is 16-17 
write offset is 18-19 
capture data write size is 20-21 
capture source is 24-25 
capture source offset is 26-27 
29-30 does something but nobody knows what yet 
31 is enable 
29-30 seems to have something to do with the blending 
blending only seems to work when 29-30 is 10 or 11 
when it's 00 or 01 only one of the two sources shows 
i can't get 10 and 11 to do anything different 
but it doesn't shut down on 11, where most things with a prohibited mode do,
so it's probably two distinct things, i just don't know what
*/
void SetRegCapture(bool enable, uint8 mode, uint8 srcBlend, uint8 destBlend, uint8 bank, uint8 offset, uint8 size, uint8 source, uint8 srcOffset)
{
   uint32 value = 0;
   if (enable)
      value |= 1 << 31; // 31 is enable
   value |= (mode & 0x3) << 29; // 29-30 seems to have something to do with the blending
   value |= (srcOffset & 0x3) << 26; // capture source offset is 26-27
   value |= (source & 0x3) << 24; // capture source is 24-25
   value |= (size & 0x3) << 20; // capture data write size is 20-21
   value |= (offset & 0x3) << 18; // write offset is 18-19 
   value |= (bank & 0x3) << 16; // vram bank select is 16-17 
   value |= (srcBlend & 0xF) << 8; // graphics blend evb is 8..12
   value |= (destBlend & 0xF) << 0; // ram blend EVA is bits 0..4 
   
   REG_CAPTURE = value;
}

int LoadGLTextures() // Load PCX files And Convert To Textures
{
	sImage pcx;                //////////////(NEW) and different from nehe.

	//load our texture
	loadPCX((u8*)ship1_texture_pcx, &pcx);
	
	image8to16(&pcx);

	glGenTextures(1, &texture[0]);
	glBindTexture(0, texture[0]);
	glTexImage2D(0, 0, GL_RGB, TEXTURE_SIZE_128 , TEXTURE_SIZE_128, 0, TEXGEN_TEXCOORD, pcx.data8);

	imageDestroy(&pcx);

	return TRUE;
}

int main()
{	
	// Turn on everything
	powerON(POWER_ALL);
	lcdSwap();
	
	// Setup the Main screen for 3D 
	videoSetMode(MODE_5_3D | DISPLAY_BG2_ACTIVE | DISPLAY_BG3_ACTIVE);
    
	// Set up the 16bit bitmap display to display the captured/processed frame
	BG2_CR = (BG_BMP16_256x256 | BG_BMP_BASE(0));
    BG2_XDX = 256;
    BG2_XDY = 0;
    BG2_YDX = 0;
    BG2_YDY = 256;
    BG2_CY = 0;
    BG2_CX = 0;
  
	// Set up the 8bit bitmap for the half-resolution glow overlay
    BG3_CR = (BG_BMP8_128x128 | BG_BMP_BASE(7));
    BG3_XDX = 128; 
    BG3_XDY = 0;
    BG3_YDX = 0;
    BG3_YDY = 128; 
    BG3_CY = 0;
    BG3_CX = 0; 

	// Put 'em in the right order
    BG0_CR |= BG_PRIORITY_3;
    BG2_CR |= BG_PRIORITY_2;
    BG3_CR |= BG_PRIORITY_1;

	// Default bank setup       
    vramSetMainBanks(VRAM_A_TEXTURE, VRAM_B_MAIN_BG_0x6000000, VRAM_C_LCD, VRAM_D_LCD);


	// Subscreen as a console
	videoSetModeSub(MODE_0_2D | DISPLAY_BG0_ACTIVE);
	vramSetBankH(VRAM_H_SUB_BG); 
	SUB_BG0_CR = BG_MAP_BASE(15);
	consoleInitDefault((u16*)SCREEN_BASE_BLOCK_SUB(15), (u16*)CHAR_BASE_BLOCK_SUB(0), 16);
	BG_PALETTE_SUB[0]=0;	
	BG_PALETTE_SUB[255]=0xff00;
	iprintf("\n\n  bluescrn's 60fps bloom demo\n");
	iprintf("  \n  who needs pixel shaders? :)\n");

	

   	// Set up a palette for the glow. The useful bit is only the first 32 colours or so, 
	// but we allow the output to overflow into the unused palette range, and be clamped
	// by the palette. This saves some per-pixel range checking/clamping
	for(int i=0;i<256;i++)
	{		
		int ri=i>>0; 
		if (ri>31) ri=31;
		if (ri<0 ) ri=0;
		int gi=i>>1; 
		if (gi>31) gi=31;
		if (gi<0 ) gi=0;
		int bi=i>>2; 
		if (bi>31) bi=31;
		if (bi<0 ) bi=0;
		BG_PALETTE[i]=0x8000+(ri*0x400+gi*0x20+bi);
	}





	// IRQ basic setup
	irqInit();
	irqSet(IRQ_VBLANK, 0);

	// Set our viewport to be the same size as the screen
	glViewPort(0,0,255,191);
	
	// Specify the Clear Color and Depth 
	glClearColor(0,0,0);
	glClearDepth(0x7FFF);
	
	LoadGLTextures();
   
	modelPtr = (ModelHeader*) ship1_model_bin;
	modelPtr->bone_ofs = (int*)((byte*) modelPtr + (int)modelPtr->bone_ofs);
	modelPtr->dlist_ofs = (int*)((byte*) modelPtr + (int)modelPtr->dlist_ofs);
   
	objectMatrix.m[0] = objectMatrix.m[5] = objectMatrix.m[10] = objectMatrix.m[15] = floatof32(1.0f);
	objectMatrix.m[1] = objectMatrix.m[2] = objectMatrix.m[3] = floatof32(0.0f);
	objectMatrix.m[4] = objectMatrix.m[6] = objectMatrix.m[7] = floatof32(0.0f);
	objectMatrix.m[8] = objectMatrix.m[9] = objectMatrix.m[11] = floatof32(0.0f);
	objectMatrix.m[12] = objectMatrix.m[13] = objectMatrix.m[14] = floatof32(0.0f);

	uint32 frameCount = 0;


	// Alloc 2 half-screen sized buffers, with added 8-pixel borders, for the blur effect
	u8 *pBlurBuffers[2];
	for(int n=0;n<2;n++)
	{
		int size=(128+16)*(96+16);
		pBlurBuffers[n]=(u8 *)malloc(size);
		memset(pBlurBuffers[n],0,size);

		// frig... set the pointers to the actual (0,0) in the buffer....
		pBlurBuffers[n]+=((128+16)*8)+8;
	}	


	while (1) 
	{
		// Reset the screen and setup the view
		glReset();
		scanKeys();

		gluPerspective(35, 256.0 / 192.0, 0.1, 100);
	
		glColor3f(1,1,1);
		
		glLight(0, RGB15(15,31,31) , 0,				  floatov10(1.0),		 0);
		glLight(1, RGB15(31,15,31) , 0,				  0,	floatov10(1.0));
		glLight(2, RGB15(31,15,15) , 0,				  0,	floatov10(1.0));
	
		glPushMatrix();
		
		glMatrixMode(GL_TEXTURE);
		glIdentity();
		
		glMatrixMode(GL_MODELVIEW);

		//need to set up some material properties since DS does not have them set by default
		glMaterialf(GL_AMBIENT, RGB15( 8, 8, 8));
		glMaterialf(GL_DIFFUSE, RGB15(15,15,15));
		glMaterialf(GL_SPECULAR, BIT(15) | RGB15(0,0,0));
		glMaterialf(GL_EMISSION, RGB15(0,0,0));

		//ds uses a table for shinyness..this generates a half-ass one
		glMaterialShinyness();

		glPolyFmt(POLY_ALPHA(31) | POLY_CULL_BACK  | POLY_FORMAT_LIGHT0| POLY_FORMAT_LIGHT1| POLY_FORMAT_LIGHT2);

		// Set the current matrix to be the model matrix
		glMatrixMode(GL_MODELVIEW);
		
		//Push our original Matrix onto the stack (save state)
		glPushMatrix();
      
		// Might as well have some edge AA - it seems free?
		glEnable(GL_ANTIALIAS);


		// Triple-buffer our VRAM banks, for 60fps postprocessing. Note that the final
		// output will lag 2 frames behind the game logic
		int tripleBuffer= frameCount%3;
		u16 *pProcessBuffer=0;

		if (tripleBuffer==0)
		{
			vramSetBankB(VRAM_B_MAIN_BG_0x6000000);		// Displaying
			vramSetBankC(VRAM_C_LCD);         				// Capturing
			vramSetBankD(VRAM_D_LCD);			 			// Processing
			SetRegCapture(true, 0, 15, 0, 2, 0, 3, 1, 0);  // Capture from C (2)
			pProcessBuffer=VRAM_D;
		}

		if (tripleBuffer==1)	
		{
			vramSetBankD(VRAM_D_MAIN_BG_0x6000000);			// Displaying
			vramSetBankB(VRAM_B_LCD);						// Capturing
			vramSetBankC(VRAM_C_LCD);			     		// Processing
			SetRegCapture(true, 0, 15, 0, 1, 0, 3, 1, 0);    
			pProcessBuffer=VRAM_C;
		}

		if (tripleBuffer==2)
		{
			vramSetBankC(VRAM_C_MAIN_BG_0x6000000);		// Displaying
			vramSetBankD(VRAM_D_LCD);						// Capturing
			vramSetBankB(VRAM_B_LCD);			     		// Processing
			SetRegCapture(true, 0, 15, 0, 3, 0, 3, 1, 0);    
			pProcessBuffer=VRAM_B;
		}





		// Submit the 3D scene now
		glClearColor(0,0,0);
		DrawShip();
  		glPopMatrix(1);


		
		// Now whilst that is rendered, we can spend the rest of the frame processing the frame at 0x6020000 

		// Grab half-screen-sized buffer for blurring
		u16 *pVram=pProcessBuffer;
		u8  *pBuffer=pBlurBuffers[0];

		// No real attempts to optimize this yet:
		for(int y=0;y<192;y=y+2)	
		{
			for(int x=0;x<256;x=x+2)
			{
				u16 val=*pVram;
				s8 v = ((val&0x1f)+((val>>5)&0x1f)+((val>>10)&0x1f)); //-10;	 // The 10 is a 'min brightness to bloom' value
				//if (v<0) v=0;
				*pBuffer++=v;			
				pVram+=2;
			}
			pBuffer+=16;	// skip over the borders
			pVram+=256;
		}




		// See here for nice explanation of how to
		// do a very fast blur filter: http://www.filtermeister.com/tutorials/blur01.html

		u8 *pSrc=0;
		u8 *pDst=0;
		int ks=5;     // ks = filter box size. Max = 8 (without increasing the borders)

		// H-blur
		pSrc=pBlurBuffers[0];
		pDst=pBlurBuffers[1];

		for(int y=0;y<96;y++)
		{
			int v=0;
			
			for(int n=-ks;n<=ks;n++)
			{
				v+=pSrc[n];
			}

			for(int x=0;x<128;x++)
			{			
				*pDst++ = v>>3;				// Adjust this shift to roughly by a divide by ks!
				v=v-pSrc[-ks];
				pSrc++;
				v=v+pSrc[ks];
				
			}
			pSrc+=16;
			pDst+=16;
		}
	
		// V-blur
		for(int x=0;x<128;x++)
		{
			const int s=144;				// stride of the buffer, in bytes (8+128+8 with borders)
			int v=0;		
		
			pSrc=pBlurBuffers[1]+x;
			pDst=pBlurBuffers[0]+x;
		
			for(int n=-ks;n<=ks;n++)
			{
				v+=pSrc[s*n];
			}
		
			for(int y=0;y<96;y++)
			{
				*pDst = v>>3;				// Adjust this shift to roughly by a divide by ks!
				pDst +=s;
				v=v-pSrc[-ks*s];			
				pSrc +=s;
				v=v+pSrc[ks*s];			
			}
		}



		// Write it back to 8bpp layer
		//
		// Todo: Eliminate this step, by making the v-blur process 2 or 4 pixels at a time, and write them
		//       directly to the glow buffer in VRAM :)
		//
		u32 *pGlowVram=(u32 *)pProcessBuffer;
		pGlowVram+=((7*16384)>>2);
		u32 *pSrcBuffer=(u32 *)pBlurBuffers[0];
		for(int y=0;y<96;y++)
		{
			for(int x=0;x<32;x++)
			{
				*pGlowVram++=*pSrcBuffer++; 
			}

			// skip border
			pSrcBuffer+=4;
		}


		if(!(keysHeld() & KEY_R))
		{	
			 // Additive blend (1.0 * BG2 + 1.0 * BG3)
		     BLEND_CR = ((8<<0)+(4<<8)+(1<<6));
	   	     BLEND_AB = (31 + (31<<8));
		}
		else
		{	
			 // turn off blending, show the top layer alone
	         BLEND_CR = 0;
		     BLEND_AB = 0;
		}
	

	 // Flush to screen	
	 glFlush();
	
      frameCount++;
	}
	
	return 0;
}

int DrawShip()
{
	ModelBone* bonePtr = (ModelBone*) modelPtr->bone_ofs;
	for (int b = 0; b < modelPtr->bone_cnt; b++) {
//		glLoadMatrix4x4(&matrixPtr[b]);

/*
      objectMatrix.m[14] = floatof32(3.0f);
		glMultMatrix4x4(&objectMatrix);
      objectMatrix.m[14] = 0;
*/
		
      if ( bonePtr[b].parent >= 0 ) {
         glRestoreMatrix(bonePtr[b].parent);
      } else {
         objectMatrix.m[14] = floatof32(-2.0f);
         glLoadMatrix4x4(&objectMatrix);
         objectMatrix.m[14] = 0;

         f32 sine = SIN[xrot &  LUT_MASK];
         f32 cosine = COS[xrot & LUT_MASK];
         objectMatrix.m[5] = cosine;      
         objectMatrix.m[6] = -sine;      
         objectMatrix.m[9] = sine;      
         objectMatrix.m[10] = cosine;      
         glMultMatrix4x4(&objectMatrix);
         objectMatrix.m[5] = floatof32(1.0f);      
         objectMatrix.m[6] = floatof32(0.0f);      
         objectMatrix.m[9] = floatof32(0.0f);
         objectMatrix.m[10] = floatof32(1.0f);
      }    

		glMultMatrix4x4(&(bonePtr[b].matrix));
      
		glStoreMatrix(b);
	}

	glBindTexture(GL_TEXTURE_2D, texture[0]);

   glCallList((u32*) modelPtr->dlist_ofs);	

	xrot+=2;
	return TRUE;			
}


